#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : model.py
# @Software: orm
# @Function:
import datetime
from typing import Dict, List, Tuple, Union

from zawn_orm.database import Database, Operation
from zawn_orm.field import BaseField, IntegerField, ObjectField
from zawn_orm.state import BaseState
from zawn_orm.tool import get_table_name, MakeID


class Model(object):
    """ 模型基类，门面模式 """

    root_field_dict: Dict[str, ObjectField] = {}
    database: Database = Database()

    v = IntegerField()  # 版本号

    def __new__(cls, *args, **kwargs):
        """ 每个类有独立的单例对象 """
        if cls.__name__ not in model_mapping:
            instance = super().__new__(cls, **kwargs)
            instance.init()
            model_mapping[instance.table_name] = model_mapping[cls.__name__] = instance
        return model_mapping[cls.__name__]

    def init(self):
        self.index_flag: bool = False  # 索引标识，为true则已经初始化了
        self.id_field: str = ''  # 唯一字段名称
        self.id_prefix: str = ''  # 唯一字段前缀
        self.table_name: str = ''  # 表名
        self.table_prefix: str = ''  # 表名前缀
        self.state = BaseState()  # 有限状态机
        self.init_table_name()  # 初始化模型的表名

    @classmethod
    def set_database(cls, database: Database):
        """ 设置数据库 """
        cls.database = database.init(cls)

    @classmethod
    def load_fields(cls) -> Dict[str, BaseField]:
        """ 读取定义的字段信息，返回字段定义字典 """
        field_dict = {}
        for field_name, field_class in cls.__dict__.items():
            if not isinstance(field_class, BaseField):
                continue
            field_dict[field_name] = field_class
        return field_dict

    @classmethod
    def objects(cls) -> BaseField:
        """ 获取字段对象 """
        if cls.__name__ not in cls.root_field_dict:
            cls.root_field_dict[cls.__name__] = ObjectField(cls.load_fields())
        return cls.root_field_dict[cls.__name__]

    def init_table_name(self, table_name: str = ''):
        """ 初始化模型的表名 """

        if self.table_name:
            return

        if table_name and isinstance(table_name, str):
            self.table_name = table_name
            return

        self.id_prefix, self.table_name = get_table_name(self.__class__)
        if self.table_prefix:
            self.table_name = '_'.join([self.table_prefix, self.table_name])

    def load(self, data: dict) -> dict:
        """ 装载数据 """
        return self.objects().load(data)

    def to_json(self, data: dict) -> dict:
        """ 转为json """
        return self.objects().to_json(data)

    def to_db(self, data: dict) -> dict:
        """ 转为数据库数据 """
        return self.objects().to_db(data)

    def new_id(self, **kwargs) -> str:
        """ 获取一个新的id """
        return MakeID.next(self.id_prefix)

    def new_operation(self, operation: str, filter: dict, update: dict) -> Operation:
        """ 获取一个新的操作 """
        return self.database.new_operation(self.table_name).set_operation(operation, filter, update)

    async def find_one(self, filter: dict, raw: bool = False, **kwargs) -> dict:
        query = self.database.new_query(self.table_name).set_query(filter)
        data = await self.database.find_one(query, **kwargs)
        if raw:
            return data
        return self.load(data)

    async def search(self, filter: dict, raw: bool = False, max_length: int = 1000, **kwargs) -> Tuple[dict, list]:
        query = self.database.new_query(self.table_name).set_query(filter)
        query.max_length = max_length
        meta, data = await self.database.search(query, **kwargs)
        if raw:
            return meta, data
        data = [self.load(i) for i in data]
        return meta, data

    async def execute(self, operation_list: List[Operation], **kwargs) -> dict:
        if not self.index_flag:
            self.index_flag = True
            await self.database.init_index(self.load_index())
        return await self.database.execute(operation_list, **kwargs)

    def load_index(self):
        """ 读取索引配置 """
        operation_list: List[Operation] = []
        for name, field in self.load_fields().items():
            if field.unique:
                operation_list.append(self.new_operation('index', filter={name: 1}, update={'unique': True}))
            elif field.index:
                operation_list.append(self.new_operation('index', filter={name: 1}, update={}))
            else:
                pass
        return operation_list

    async def insert(
            self, update: dict, return_operation: bool = False, **kwargs) -> Union[List[Operation], dict]:
        """ 快捷新增 """

        # 如果没有主键字段，则自动新增一个
        if self.id_field is not None and self.id_field in update:
            update[self.id_field] = self.new_id(**update)

        # 写入数据库
        operation = self.new_operation('insert', {}, self.to_db(update))
        if return_operation:
            return [operation]
        result = await self.execute([operation], **kwargs)
        return result.get(self.table_name, {})

    async def update(
            self, filter: dict, update: dict, return_operation: bool = False, **kwargs) -> Union[List[Operation], dict]:
        """ 快捷更新 """

        operation = self.new_operation('update', filter, update)
        if return_operation:
            return [operation]
        result = await self.execute([operation], **kwargs)
        return result.get(self.table_name, {})

    async def delete(
            self, filter: dict, return_operation: bool = False, **kwargs) -> Union[List[Operation], dict]:
        """ 快捷删除 """
        now = datetime.datetime.now().astimezone()
        op = filter.get('op', 'system')
        _, data = await self.search(filter)

        operaion_list = [self.new_operation('delete', filter, {})]
        for i in data:
            i.update({
                'delete_at': now,
                'delete_by': op,
            })
            insert_operation = self.new_operation('insert', {}, i)
            insert_operation.table_name = f'__delete_{self.table_name}'
            operaion_list.append(insert_operation)

        if return_operation:
            return operaion_list
        result = await self.execute(operaion_list, **kwargs)
        return result.get(self.table_name, {})


# 模型映射表
model_mapping: Dict[str, Model] = {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : sqlite.py
# @Software: orm
# @Function:
import sqlite3
from typing import List

from zawn_orm.database import Database, Query, Operation


class SQLite(Database):
    """ SQLite数据库的实现类 """

    database_type: str = 'sqlite'

    async def connect(self):
        self.connection = sqlite3.connect(':memory:')

    async def close(self):
        self.connection.commit()
        self.connection.close()
        self.connection = None

    async def on_search(self, query: Query):
        pass

    async def on_execute(self, operation_list: List[Operation]) -> dict:
        cursor = self.connection.cursor()
        cursor.executemany()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-09-10
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : codec.py
# @Software: orm
# @Function:
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Union, Type

from zawn_orm.exception import InvalidValueException
from zawn_orm.tool import datetime_format, rounding

__all__ = ['DefaultCodec', 'CodecFactory']


class DefaultCodec(object):
    """ 默认字段数据类型的编解码器 """
    type_name = '_'
    default_value = None

    def on_transform(self, data: Any) -> Any:
        """ 默认转换方法 """
        return data

    def on_load(self, data: Any) -> Any:
        """ 装载方法 """
        return self.on_transform(data)

    def on_to_json(self, data: Any) -> Any:
        """ 输出JSON """
        return self.on_transform(data)

    def on_to_db(self, data: Any) -> Any:
        """ 输出数据库类型 """
        return self.on_transform(data)


class IDCodec(DefaultCodec):
    """ 唯一主键编解码类 """
    type_name = 'id'


class BooleanCodec(DefaultCodec):
    """ 布尔编解码类 """
    type_name = 'boolean'

    def on_transform(self, data: Any) -> bool:
        return bool(data)


class StringCodec(DefaultCodec):
    """ 字符串编解码类 """
    type_name = 'string'

    def on_transform(self, data: Any) -> str:
        return str(data)


class FloatCodec(DefaultCodec):
    """ 浮点数字编解码类 """
    type_name = 'float'

    def on_transform(self, data: Union[str, int, float, Decimal]) -> float:
        return float(data)


class IntegerCodec(DefaultCodec):
    """ 整型数字编解码类 """
    type_name = 'integer'

    def on_transform(self, data: Union[str, int, float, Decimal]) -> int:
        return int(data)


class DecimalCodec(DefaultCodec):
    """ 十进制数字编解码类 """
    type_name = 'decimal'

    def on_load(self, data: Union[str, int, float, Decimal]) -> Decimal:
        return rounding(Decimal(data), 4)

    def on_to_json(self, data: Decimal) -> str:
        return str(rounding(Decimal(data), 4))

    def on_to_db(self, data: Decimal) -> float:
        return float(data)


class DatetimeCodec(DefaultCodec):
    """ 日期时间编解码类 """
    type_name = 'datetime'

    def on_load(self, data: Union[str, int, float]) -> datetime:
        if isinstance(data, str):
            return datetime.strptime(data, datetime_format).astimezone()
        elif isinstance(data, (int, float)):
            return datetime.fromtimestamp(int(data / 1000)).astimezone()
        elif isinstance(data, datetime):
            return data.astimezone()
        raise InvalidValueException(f'{str(data)}无法转换成{self.type_name}类型')

    def on_to_json(self, data: Union[datetime]) -> str:
        if isinstance(data, datetime):
            return data.astimezone().strftime(datetime_format)
        raise InvalidValueException(f'{str(data)}无法转换成{self.type_name}类型')

    def on_to_db(self, data: Union[datetime]) -> int:
        if isinstance(data, datetime):
            return int(data.timestamp() * 1000)
        raise InvalidValueException(f'{str(data)}无法转换成{self.type_name}类型')


class CodecFactory(object):
    """ 编解码器工厂 """

    codec_tuple = (
        DefaultCodec, IDCodec, BooleanCodec,
        StringCodec, FloatCodec, IntegerCodec,
        DecimalCodec, DatetimeCodec,
    )
    codec_mapping: Dict[str, DefaultCodec] = {c.type_name: c() for c in codec_tuple}

    @classmethod
    def set(cls, codec: Type[DefaultCodec], force: bool = False):
        if codec.type_name in cls.codec_mapping and not force:
            return
        cls.codec_mapping[codec.type_name] = codec()

    @classmethod
    def get(cls, type_name: str) -> DefaultCodec:
        return cls.codec_mapping.get(type_name, cls.codec_mapping[DefaultCodec.type_name])


if __name__ == '__main__':
    test_data = (-1, 0, 1, '-', '2022-09-10T16:30:00.123456+0800', datetime.now())
    for type_name, codec in CodecFactory.codec_mapping.items():
        print('-' * 20)
        for i in test_data:
            for ii in ('on_load', 'on_to_json', 'on_to_db'):
                try:
                    method = getattr(codec, ii)
                    print(type_name, i, ii, method(i))
                except Exception as e:
                    print(type_name, i, ii, f'出错:{e}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : test.py
# @Software: orm
# @Function:
import datetime
import json
import time
import unittest

from zawn_orm import field
from zawn_orm import model
from zawn_orm.impl.mongodb import MongoDB
from zawn_orm.tool import datetime_format, MakeID

with open('../uri.json', 'rb') as fp:
    config = json.load(fp)

test_json = {
    'id': int(time.time() - 100),
    'name': MakeID.next(''),
    'height': 180.5,
    'create_time': '2022-01-01T00:00:00.000+0800',
    'balance': '100.00',
    'enabled': True,
    'object_field': {
        's_id': '11',
        's_name': 11,
        's_height': '181.5',
        's_create_time': '2020-01-01T00:00:00.000+0800',
        's_balance': '200.00',
        's_enabled': 1,
    },
    'array_field': [
        {
            's_id': '11',
            's_name': 11,
            's_height': '181.5',
            's_balance': '200.00',
            's_enabled': 1,
            'b': 123,
        },
        {
            's_id': '110',
            's_name': 110,
            's_height': '185.5',
            's_balance': '250.00',
            's_enabled': 0,
            'b': lambda x: x,
        },
    ],
}


class TestMongoDB(unittest.IsolatedAsyncioTestCase):

    async def init_database(self):
        database = MongoDB()
        database.connect(config['mongodb_uri'])
        model.Model.set_database(database)

    def init_class(self):
        class EmbeddedModel(model.Model):
            s_name = field.StringField()
            s_id = field.IntegerField()
            s_height = field.FloatField()
            s_create_time = field.DatetimeField()
            s_balance = field.DecimalField()
            s_enabled = field.BooleanField()

        class TestModel(model.Model):
            id = field.IDField(index=True)
            name = field.StringField(unique=True)
            height = field.FloatField()
            create_time = field.DatetimeField()
            balance = field.DecimalField()
            enabled = field.BooleanField()
            object_field = field.ObjectField(EmbeddedModel.load_fields())
            array_field = field.ArrayField(EmbeddedModel.load_fields())

        self.EmbeddedModel = EmbeddedModel
        self.TestModel = TestModel

    async def asyncSetUp(self) -> None:
        self.test_json = test_json
        await self.init_database()
        self.init_class()

    async def asyncTearDown(self) -> None:
        pass

    async def test__objects_equal(self):
        self.assertEqual(self.TestModel().table_name, 'test_model')
        self.assertEqual(self.EmbeddedModel().table_name, 'embedded_model')
        self.assertEqual(self.TestModel().table_name, 'test_model')
        self.assertEqual(self.TestModel(), self.TestModel(), '相同模型的objects返回对象应该相同')
        self.assertEqual(self.EmbeddedModel(), self.EmbeddedModel(), '相同模型的objects返回对象应该相同')
        self.assertNotEqual(self.TestModel(), self.EmbeddedModel(), '不同模型的objects返回对象应该不同')

    async def test__transform(self):
        now = datetime.datetime.now().astimezone()
        now_str = now.strftime(datetime_format)
        test_model = self.TestModel()
        test_json = self.test_json
        test_json['create_time'] = now_str

        data = test_model.objects().load(test_json)
        test_to_db = test_model.to_db(data)
        test_to_json = test_model.to_json(data)

        self.assertEqual(data['create_time'], now)
        self.assertEqual(test_to_db['create_time'], now)
        self.assertEqual(test_to_json['create_time'], now_str)

    async def test__write_and_read(self):
        now = datetime.datetime.now().astimezone()
        now_str = now.strftime(datetime_format)
        id_1 = int(now.timestamp())
        id_2 = id_1 + 1
        test_model = self.TestModel()
        test_json = self.test_json
        test_json['create_time'] = now_str
        test_json['id'] = id_1
        test_json['name'] = MakeID.next('')
        test_json_2 = test_json.copy()
        test_json_2['id'] = id_2
        test_json_2['name'] = MakeID.next('')

        operation_list = [
            test_model.new_operation('delete', {}, {}),
            test_model.new_operation('insert', {}, test_json),
            test_model.new_operation('update', {'id': id_1}, {'$inc': {'balance': 100}}),
            test_model.new_operation('insert', {}, test_json_2),
            test_model.new_operation('delete', {'id': id_2}, {}),
        ]
        result = await test_model.execute(operation_list)

        meta, data = await test_model.search({'$project': {'_id': 0, 'create_time': 1}})
        test_to_json = [test_model.to_json(i) for i in data]

        self.assertEqual(result[test_model.table_name]['insert'], 2)
        self.assertGreaterEqual(result[test_model.table_name]['update'], 0)
        self.assertGreater(meta['total'], 0)
        self.assertTrue(test_to_json[0]['create_time'].startswith(now_str[:10]))
        self.assertIsNone(test_to_json[0].get('id'))

    async def test__quick_write_and_read(self):
        now = datetime.datetime.now().astimezone()
        now_str = now.strftime(datetime_format)
        id_1 = int(now.timestamp()) + 2
        test_model = self.TestModel()
        test_json = self.test_json
        test_json['create_time'] = now_str
        test_json['id'] = id_1
        test_json['name'] = MakeID.next('')

        delete_result_0 = await test_model.delete({})
        insert_result = await test_model.insert(test_json, assert_dict={'test_model': {'insert': 1}})
        update_result = await test_model.update({'id': id_1}, {'name': 'set_name', '$inc': {'balance': 55}})
        one_data = await test_model.find_one({})
        meta, data = await test_model.search(
            {'$filter': [
                {'$group': {'_id': '$name', 'count': {'$sum': '$balance'}}},
            ]},
            raw=True,
        )
        delete_result_1 = await test_model.delete({}, assert_dict={'test_model': {'delete': 1}})
        delete_result_2 = await test_model.delete({}, assert_dict={'test_model': {'delete': 0}})

        self.assertIn('id', one_data)
        self.assertGreaterEqual(meta['total'], 1)
        self.assertGreaterEqual(data[0]['count'], 1)
        self.assertEqual(insert_result['insert'], 1)
        self.assertGreaterEqual(update_result['update'], 1)
        self.assertIn('delete', delete_result_1)
        self.assertIn('delete', delete_result_2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : field.py
# @Software: orm
# @Function:
import logging
from typing import Union, Dict

from zawn_orm.codec import CodecFactory

DEBUG = False


class BaseField(object):
    """ 基础字段类 """
    type_name = '_'

    def __init__(self, **kwargs):
        self.class_name = self.__class__.__name__
        self.unique: bool = bool(kwargs.get('unique'))
        self.index: bool = bool(kwargs.get('index'))

    def __str__(self):
        return f'<{self.class_name}>'

    def __repr__(self):
        return f'<{self.class_name}>'

    def load(self, data):
        """ 装载数据 """
        codec = CodecFactory.get(self.type_name)
        try:
            return codec.on_load(data)
        except Exception as e:
            if DEBUG:
                logging.error(f'字段值读取出错:{self.class_name}:{data}<-{type(data)}:{e}', exc_info=True)
            return codec.default_value

    def to_json(self, data):
        """ 转为json """
        codec = CodecFactory.get(self.type_name)
        try:
            return codec.on_to_json(data)
        except Exception as e:
            if DEBUG:
                logging.error(f'字段值转JSON出错:{self.class_name}:{data}<-{type(data)}:{e}', exc_info=True)
            return codec.default_value

    def to_db(self, data):
        """ 转为数据库数据 """
        codec = CodecFactory.get(self.type_name)
        try:
            return codec.on_to_db(data)
        except Exception as e:
            if DEBUG:
                logging.error(f'字段值转数据库格式出错:{self.class_name}:{data}<-{type(data)}:{e}', exc_info=True)
            return codec.default_value


class IDField(BaseField):
    """ 唯一主键字段 """
    type_name = 'id'


class BooleanField(BaseField):
    """ 布尔字段类 """
    type_name = 'boolean'


class StringField(BaseField):
    """ 字符串字段类 """
    type_name = 'string'


class FloatField(BaseField):
    """ 浮点数字字段类 """
    type_name = 'float'


class IntegerField(BaseField):
    """ 整型数字字段类 """
    type_name = 'integer'


class DecimalField(BaseField):
    """ 十进制数字字段类 """
    type_name = 'decimal'


class DatetimeField(BaseField):
    """ 日期时间字段类 """
    type_name = 'datetime'


class ObjectField(BaseField):
    """ 对象字段类 """
    type_name = 'object'

    def __init__(self, field_dict: Dict[str, BaseField], **kwargs):
        super().__init__(**kwargs)
        self.field_dict = field_dict

    def object_transform(self, data: dict, method_name: str) -> dict:
        out = {}
        for codec_name, codec_object in self.field_dict.items():
            value = data.get(codec_name)
            if value is None:
                continue
            method = getattr(codec_object, method_name)
            if not callable(method):
                continue
            out[codec_name] = method(value)
        return out

    def load(self, data: dict) -> dict:
        return self.object_transform(data, 'load')

    def to_json(self, data: dict) -> dict:
        return self.object_transform(data, 'to_json')

    def to_db(self, data: dict) -> dict:
        return self.object_transform(data, 'to_db')


class ArrayField(BaseField):
    """ 数组字段类 """
    type_name = 'array'

    def __init__(self, field_dict: Union[BaseField, Dict[str, BaseField]], **kwargs):
        super().__init__(**kwargs)
        self.field_dict = field_dict

    def array_transform(self, data: list, method_name: str) -> list:
        out = []
        for i in data:
            # 如果字段定义是字段类型，则直接转换存入，并且跳过循环
            if isinstance(self.field_dict, BaseField):
                method = getattr(self.field_dict, method_name)
                if not callable(method):
                    continue
                out.append(method(i))
                continue

            row = {}
            for field_name, field_class in self.field_dict.items():
                value = i.get(field_name)
                if value is None:
                    continue
                method = getattr(field_class, method_name)
                if not callable(method):
                    continue
                row[field_name] = method(value)
            out.append(row)

        return out

    def load(self, data: list) -> list:
        return self.array_transform(data, 'load')

    def to_json(self, data: list) -> list:
        return self.array_transform(data, 'to_json')

    def to_db(self, data: list) -> list:
        return self.array_transform(data, 'to_db')


field_mapping: Dict[str, BaseField] = {}

if __name__ == '__main__':
    class Model(object):

        @classmethod
        def load_fields(cls) -> Dict[str, BaseField]:
            field_dict = {}
            for field_name, field_class in cls.__dict__.items():
                if not isinstance(field_class, BaseField):
                    continue
                field_dict[field_name] = field_class
            return field_dict

        @classmethod
        def objects(cls) -> BaseField:
            return ObjectField(cls.load_fields())


    class SubModel(Model):
        s_a = StringField()
        s_b = BooleanField()
        s_c = FloatField()
        s_d = IntegerField()


    class MainModel(Model):
        a = ObjectField(SubModel.load_fields())
        b = ArrayField(SubModel.load_fields())
        c = DatetimeField()
        d = DecimalField()
        e = IntegerField()
        f = IntegerField()


    data = {
        'a': {
            's_a': 1,
            's_b': 1,
            's_c': 1,
            's_d': 1,
            's_e': 1,
        },
        'b': [
            {
                's_a': 0,
                's_b': 0,
                's_c': 0,
                's_d': 0,
                's_e': 0,
            },
            {
                's_a': -1,
                's_b': -1,
                's_c': -1,
                's_d': -1,
                's_e': -1,
            },
        ],
        'c': '2022-08-15T16:00:30.000+0800',
        'd': '2.50',
        'e': '-',
    }

    print('')
    print(data)
    data_class = MainModel.objects()

    print('---', '装载数据')
    new_data = data_class.load(data)
    print(new_data)

    print('---', '输出JSON')
    json_data = data_class.to_json(new_data)
    print(json_data)

    print('---', '输出DB')
    db_data = data_class.to_db(new_data)
    print(db_data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : database.py
# @Software: orm
# @Function:
import logging
from typing import List, Dict, Tuple

from zawn_orm.exception import InvalidOperationException

operation_tuple = ('insert', 'update', 'delete', 'index')


class Query(object):
    """ 数据库查询类 """

    max_length: int = 1000  # 查询时最大行数限制

    def __init__(self, table_name: str):
        self.table_name = table_name
        self.filter = {}

    def set_query(self, filter: dict) -> 'Query':
        """ 设置查询条件 """
        self.filter = filter
        return self

    def get_query(self):
        """ 获取查询条件 """
        return self.filter


class Operation(object):
    """ 数据库操作类 """

    def __init__(self, table_name: str):
        self.table_name = table_name
        self.operation = None
        self.filter = {}
        self.update = {}

    def set_operation(self, operation: str, filter: dict, update: dict) -> 'Operation':
        """ 设置操作属性 """

        if operation not in operation_tuple:
            raise InvalidOperationException(f'{operation}是无效操作')

        self.operation = operation
        self.filter = filter
        self.update = update

        return self

    def get_operation(self):
        return self.table_name, self.operation, self.filter, self.update


class Result(object):
    """ 数据库操作结果 """

    def __init__(self):
        self.result_dict = {}

    def add(self, table_name: str, operation: str, number: int = 1) -> int:
        """ 增加操作统计数 """

        if table_name not in self.result_dict:
            self.result_dict[table_name] = {i: 0 for i in operation_tuple}

        if operation not in operation_tuple:
            return 0

        self.result_dict[table_name][operation] += number
        return self.result_dict[table_name][operation]

    def get(self) -> Dict[str, dict]:
        return self.result_dict


class Database(object):
    """ 数据库基类 """

    singleton_mapping: Dict[str, 'Database'] = {}
    database_type: str = 'base'
    client = None
    database = None

    class OperationCount(object):
        insert: int = 0
        update: int = 0
        delete: int = 0

        def result(self) -> dict:
            return {
                'insert': self.insert,
                'update': self.update,
                'delete': self.delete,
            }

        def __str__(self) -> str:
            return str(self.result())

        def __repr__(self) -> str:
            return str(self.result())

    def __new__(cls, *args, **kwargs):
        """ 每个类有独立的单例对象 """
        instance = cls.singleton_mapping.get(cls.__name__)
        if instance is None:
            instance = super().__new__(cls, **kwargs)
            cls.singleton_mapping[cls.__name__] = instance
        return instance

    def __init__(self):
        pass

    def init(self, model_class, **kwargs):
        return self

    async def connect(self, uri: str) -> 'Database':
        return self

    async def close(self):
        pass

    def new_query(self, table_name: str) -> Query:
        """ 获取一个新的查询 """
        return Query(table_name)

    def new_operation(self, table_name: str) -> Operation:
        """ 获取一个新的操作 """
        operation = Operation(table_name)
        return operation

    async def find_one(self, query: Query, **kwargs) -> dict:
        try:
            return await self.on_find_one(query, **kwargs)
        except Exception as e:
            logging.error(f'数据库执行查询数据单条时出错{e}', exc_info=True)
            return {}

    async def search(self, query: Query, **kwargs) -> Tuple[dict, list]:
        try:
            return await self.on_search(query, **kwargs)
        except Exception as e:
            logging.error(f'数据库执行查询时出错{e}', exc_info=True)
            return {'skip': 0, 'limit': 0, 'total': 0}, []

    async def execute(self, operation_list: List[Operation], **kwargs) -> dict:
        result = {}
        try:
            result = await self.on_execute(operation_list, **kwargs)
        except Exception as e:
            logging.error(f'数据库执行变更时出错{e}', exc_info=True)
        return result

    async def on_find_one(self, query: Query, **kwargs) -> dict:
        return {}

    async def on_search(self, query: Query, **kwargs) -> Tuple[dict, list]:
        return {'skip': 0, 'limit': 0, 'total': 0}, []

    async def on_execute(self, operation_list: List[Operation], **kwargs) -> dict:
        result = Result()
        for i in operation_list:
            result.add(i.table_name, i.operation, 1)
        return result.get()

    async def init_index(self, operation_list: List[Operation]) -> dict:
        """ 初始化索引 """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : exception.py
# @Software: orm
# @Function:


class EmptyModelException(Exception):
    """ 空模型异常类 """


class InvalidFieldException(Exception):
    """ 无效字段异常类 """


class ExisitFieldException(Exception):
    """ 存在字段异常类 """


class InvalidValueException(Exception):
    """ 无效值异常类 """


class InvalidOperationException(Exception):
    """ 无效操作异常类 """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-25
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : tool.py
# @Software: orm
# @Function:
import os
import socket
import time
from decimal import Decimal, ROUND_HALF_UP
from typing import Union

datetime_format = '%Y-%m-%dT%H:%M:%S.%f%z'


def get_table_name(model_class):
    """ 获取模型类的名字，转换成表名 """

    prefix_list = []
    chat_list = []
    for i in model_class.__name__:
        if i == i.upper():
            prefix_list.append(i)  # 大写前缀
            chat_list.append('_')
            chat_list.append(i.lower())
        else:
            chat_list.append(i)
    if chat_list[0] == '_':
        chat_list.pop(0)
    return ''.join(prefix_list), ''.join(chat_list)


def get_machine_number(mod_number=65536):
    """ 根据主机名称和进程号生成机器号码 """
    hostname = socket.gethostname().encode('utf-8')
    machine_number = os.getpid()
    for i in hostname:
        machine_number = ((machine_number << 8) + i) % mod_number
    return machine_number


class MakeID(object):
    '''生成code'''
    sn = int(time.time()) % 65536
    maps = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    length = len(maps)
    mod_number = length ** 2
    pid = get_machine_number(length ** 4)

    @classmethod
    def mod(cls, num, length):
        ids = []
        while num >= length:
            num, y = divmod(num, length)
            ids.append(cls.maps[y])
        ids.append(cls.maps[num])
        id = ''.join(ids[::-1])
        return id

    @classmethod
    def next(cls, prefix='', code_length=16, size_type: str = 'long'):
        cls.sn = (cls.sn + 1) % cls.mod_number
        fid = cls.maps[0] * code_length
        tid = f'{cls.mod(int(time.time() * 1000), cls.length):0>6}'
        pid = f'{cls.mod(cls.pid, cls.length):0>4}'
        sid = f'{cls.mod(cls.sn, cls.length):0>2}'
        id = (fid + tid + pid + sid)[-code_length + len(prefix):]
        code = prefix + id
        return code


def rounding(number: Union[int, float, str, Decimal], dig: int = 4):
    """ 对数字进行四舍五入 """

    n = Decimal(number)
    dig = int(dig)
    return n.quantize(
        Decimal('0' if dig <= 0 else f'0.{"0" * dig}'),
        rounding=ROUND_HALF_UP,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-27
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : mongodb.py
# @Software: orm
# @Function:
import logging
import re
from datetime import datetime, timedelta
from decimal import Decimal
from typing import List, Tuple, Any, Union

from bson import codec_options, ObjectId, decimal128
from motor.core import AgnosticCollection
from motor.motor_asyncio import AsyncIOMotorClient
from pymongo import InsertOne, UpdateMany, DeleteMany, IndexModel

from zawn_orm.codec import DefaultCodec, CodecFactory
from zawn_orm.database import Database, Query, Operation
from zawn_orm.exception import InvalidOperationException, InvalidValueException
from zawn_orm.model import model_mapping
from zawn_orm.tool import datetime_format, rounding


class SearchUtils(object):
    """ 列表搜索基类，模版模式 """

    key_regex = re.compile('^[0-9a-zA-Z_]+$')
    default_skip: int = 0
    default_limit: int = 10

    def __init__(self, collection):
        self.collection = collection

    def match(self, conditions: List[dict]):
        query = [{'$match': {'$and': conditions}}]
        return query

    def sort_desc_by_create_time(self):
        query = [{'$sort': {'create_at': -1, '_id': -1}}]
        return query

    def skip_limit(self, skip: int = default_skip, limit: int = default_limit):
        query = []
        if skip > 0:
            query.append({'$skip': skip})
        if limit > 0:
            query.append({'$limit': limit})
        return query

    def unwind(self, path: str, preserve: bool = True):
        query = [{'$unwind': {'path': path, 'preserveNullAndEmptyArrays': preserve}}]
        return query

    def join(
            self, from_collection: str, local_field: str, foreign_field: str,
            as_name: str, unwind_path: str = None, set_fields: dict = None,
            preserve: bool = True):
        query = [
            {'$lookup': {
                'from': from_collection,
                'localField': local_field,
                'foreignField': foreign_field,
                'as': as_name,
            }},
        ]
        # 这里改为最后参数unwind_path不是None时，才增加$unwind语句
        if unwind_path is not None:
            query.extend(self.unwind(unwind_path, preserve))
        if set_fields:
            query.extend([
                {'$set': {
                    i: f'${as_name}.{set_fields[i]}' for i in set_fields
                }},
                {'$unset': [as_name]},
            ])
        return query

    def group_by(self, fields: [dict, str], count: dict):
        ''' 分组统计 '''

        query = []
        group_query = {'$group': {
            '_id': fields,
            'root': {'$mergeObjects': '$$ROOT'},
        }}
        group_query['$group'].update(count)
        query.append(group_query)

        if len(count) != 0:
            query.append({'$set': {f'root.{i}': f'${i}' for i in count}})
        query.append({'$replaceRoot': {'newRoot': '$root'}})

        return query

    def join2(self, from_collection, join_fields, set_fields):
        ''' 联表查询语句，多字段联表
        set_fields={原表字段名: 关联表字段名}

        *QueryUtils.join2(
            'aid_train_result',
            {'train_code': 'train_code', 'aid_user_code': 'aid_user_code'},
            {
                'theory_result': 'theory_result',
                'trial_result': 'trial_result',
                'recovery_result': 'recovery_result',
                'rescue_result': 'rescue_result',
                'attendance': 'attendance',
            },
        ),
        '''

        let_fields = {}
        conditions = []
        for k in join_fields:
            let_fields[k] = f'${k}'
            conditions.append({'$eq': [f'${join_fields[k]}', f'$${k}']})
        conditions.append({'$ne': ['$status', '-1']})

        for k in set_fields:
            set_fields[k] = {'$first': f'${from_collection}.{set_fields[k]}'}

        pipeline = [
            {'$lookup': {
                'from': from_collection,
                'let': let_fields,
                'pipeline': [
                    {'$match': {'$expr': {'$and': conditions}}},
                ],
                'as': from_collection,
            }},
            {'$set': set_fields},
            {'$unset': [from_collection]},
        ]

        return pipeline

    def range_query(
            self, key: str, value: [list, str],
            query_type: str, offset_time: timedelta = timedelta()):
        ''' 生成返回搜索条件
        搜索类型 query_type:[date,datetime,number,code]
        '''
        query = []
        if query_type == 'date':
            date_format = '%Y-%m-%d'
            one_day = timedelta(days=1)
            if isinstance(value, list) and len(value) == 2:
                start = datetime.strptime(value[0], date_format) + offset_time
                end = datetime.strptime(value[1], date_format) + offset_time + one_day
                query.append({key: {'$gte': start, '$lt': end}})
            elif isinstance(value, str):
                dt = datetime.strptime(value, date_format) + offset_time
                query.append({key: {'$gte': dt, '$lt': dt + one_day}})
            else:
                raise Exception(f'时间搜索条件 {key}:[start,end]')

        elif query_type == 'datetime':
            datetime_format = '%Y-%m-%dT%H:%M:%S.%fZ'
            if isinstance(value, list) and len(value) == 2:
                start = datetime.strptime(value[0], datetime_format)
                end = datetime.strptime(value[1], datetime_format)
                query.append({key: {'$gte': start, '$lt': end}})
            elif isinstance(value, str):
                query.append({key: datetime.strptime(value, datetime_format)})
            else:
                raise Exception(f'时间搜索条件 {key}:[start,end]')

        elif query_type == 'number':
            if isinstance(value, list) and len(value) == 2:
                query.append({key: {'$gte': value[0], '$lte': value[1]}})
            elif isinstance(value, (int, float)):
                query.append({key: value})
            elif isinstance(value, str):
                query.append({key: float(value)})
            else:
                raise Exception(f'数字搜索条件 {key}:[start,end]')

        elif query_type == 'code':
            if isinstance(value, list):
                query.append({key: {'$in': value}})
            elif isinstance(value, str):
                query.append({key: value})
            else:
                raise Exception(f'数字搜索条件 {key}:[code1,code2,...]')

        elif query_type == 'status':
            if isinstance(value, list):
                query.append({key: {'$in': value}})
            elif isinstance(value, str):
                query.append({key: value})
            else:
                query.append({key: {'$ne': '-1'}})

        else:
            raise Exception('必传query_type:[date,datetime,number,code]')

        return query

    def add_filter(self, raw_query: dict) -> List[dict]:
        """ 加入过滤条件 """

        pipeline = []

        query = {'status': raw_query.pop('status', {'$nin': ['S3', 'S8']})}

        for key, value in raw_query.items():
            if not self.key_regex.match(key):
                continue
            query[key] = value

        filter_or = raw_query.get('$or')
        if isinstance(filter_or, list):
            query['$or'] = filter_or

        pipeline.append({'$match': query})

        return pipeline

    def add_paging(self, raw_query: dict) -> List[dict]:
        """ 加入分页条件 """

        pipeline = []

        sort = raw_query.get('$sort')
        skip = raw_query.get('$skip') or self.default_skip
        limit = raw_query.get('$limit') or self.default_limit

        pipeline.append({'$sort': (sort if sort else {'create_at': -1})})
        if skip > 0:
            pipeline.append({'$skip': skip})
        if limit > 0:
            pipeline.append({'$limit': limit})

        return pipeline

    def add_output(self, raw_query: dict) -> List[dict]:
        """ 加入输出条件 """

        pipeline = []

        project = raw_query.get('$project')
        pipeline.append({'$project': (project if project else {'_id': 0})})

        return pipeline

    async def find_one(self, raw_query: dict) -> Tuple[list, dict]:
        """ 执行查找方法 """

        pipeline, data = await self.search(raw_query)

        return pipeline, (data and data[0] or {})

    async def search(self, raw_query: dict) -> Tuple[List[dict], List[dict]]:
        """ 执行搜索方法 """

        append_pipeline = raw_query.get('$pipeline') or []

        pipeline = [
            *self.add_filter(raw_query),  # 过滤条件
            *self.add_paging(raw_query),  # 排序分页
            *append_pipeline,  # 其他操作
            *self.add_output(raw_query),  # 输出字段
        ]

        data = await self.collection.aggregate(pipeline)

        return pipeline, data

    async def count(self, raw_query: dict) -> dict:
        """ 执行统计方法 """

        skip = raw_query.get('$skip') or self.default_skip
        skip = int(skip) if skip > 0 else 0

        limit = raw_query.get('$limit') or self.default_limit
        limit = int(limit) if limit > 0 else 0

        pipeline = [
            *self.add_filter(raw_query),
            {'$count': 'total'},
        ]
        data = await self.collection.aggregate(pipeline)
        total = data and data[0].get('total', 0) or 0

        return {'skip': skip, 'limit': limit, 'total': total}


class MongoDBQuery(Query):
    """ MongoDB查询类 """

    default_limit: int = 20

    def paging(self, pragma_dict: dict) -> Tuple[dict, int, int]:
        """ 获取分页条件 """

        sort = pragma_dict.get('$sort') or {}
        if not isinstance(sort, dict):
            sort = {}

        skip = pragma_dict.get('$skip') or 0
        if isinstance(skip, int):
            skip = 0 if skip <= 0 else skip
        else:
            skip = 0

        # limit为0时不限制，大于0时增加限制，其他默认20行限制
        limit = pragma_dict.get('$limit') or self.default_limit
        if isinstance(limit, int):
            limit = 0 if limit <= 0 else limit
        else:
            limit = self.default_limit

        return sort, skip, limit

    def get_query(self) -> Tuple[str, list, list, int, int]:
        """ 获取查询条件 """

        new_filter = self.filter.copy()
        pragma_dict = {}

        count_pipeline = []
        data_pipeline = []

        for k, v in self.filter.items():
            if not isinstance(k, str):
                continue
            if k.startswith('$') and k not in ['$and', '$or']:
                pragma_dict[k] = new_filter.pop(k)

        # 需要在分页前的过滤阶段
        filter = pragma_dict.get('$filter')
        if not isinstance(filter, list):
            filter = []

        count_pipeline.extend([
            {'$match': new_filter},
            *filter,
            {'$count': 'count'},
        ])

        # match阶段
        data_pipeline.extend([{'$match': new_filter}, *filter])

        # 分页
        sort, skip, limit = self.paging(pragma_dict)
        if sort:
            data_pipeline.append({'$sort': sort})
        if skip:
            data_pipeline.append({'$skip': skip})
        if limit:
            data_pipeline.append({'$limit': limit})

        pragma_handler_list = (
            ('$pipeline', lambda x: x if isinstance(x, list) else []),
            ('$unset', lambda x: [{'$unset': x}] if isinstance(x, list) else []),
            ('$project', lambda x: [{'$project': x}] if isinstance(x, dict) else []),
        )
        for k, handler in pragma_handler_list:
            value = pragma_dict.get(k)
            if value is None:
                continue
            stage = handler(value)
            data_pipeline.extend(stage)

        return self.table_name, count_pipeline, data_pipeline, skip, limit


class MongoDBOperation(Operation):
    """ MongoDB操作类 """

    def get_operation(self) -> Tuple[str, Any]:
        """ 获取操作对象 """

        # 操作相关的模型
        model = model_mapping.get(self.table_name)

        if self.operation == 'insert':
            update = self.update
            if model:
                update = model.to_db(update)
            operation_object = InsertOne(document=update)

        elif self.operation == 'update':
            # 其他操作类型的字段值
            update = {k: v for k, v in self.update.items() if k.startswith('$')}
            # 直接设置到字段值
            set_update = {k: v for k, v in self.update.items() if not k.startswith('$')}
            if model:
                set_update = model.to_db(set_update)
            if set_update:
                update['$set'] = set_update

            operation_object = UpdateMany(filter=self.filter, update=update)

        elif self.operation == 'delete':
            operation_object = DeleteMany(filter=self.filter)

        elif self.operation == 'index':
            keys = [(k, v) for k, v in self.filter.items()]
            operation_object = IndexModel(keys=keys, **self.update)

        else:
            raise InvalidOperationException(f'{self.operation}是无效操作')

        return self.table_name, operation_object


class MongoDBDecimalCodec(codec_options.TypeCodec):
    python_type = Decimal
    bson_type = decimal128.Decimal128

    def transform_python(self, value):
        return decimal128.Decimal128(value)

    def transform_bson(self, value):
        return value.to_decimal()


class IDCodec(DefaultCodec):
    """ 唯一主键编解码类 """
    type_name = 'id'

    def on_load(self, data: str) -> ObjectId:
        return ObjectId(data)

    def on_to_json(self, data: ObjectId) -> str:
        return str(data)

    def on_to_db(self, data: ObjectId) -> ObjectId:
        return data


class DecimalCodec(DefaultCodec):
    """ 十进制数字编解码类 """
    type_name = 'decimal'

    def on_load(self, data: Union[str, int, float, Decimal]) -> Decimal:
        return rounding(Decimal(data), 4)

    def on_to_json(self, data: Decimal) -> str:
        return str(rounding(Decimal(data), 4))

    def on_to_db(self, data: Decimal) -> Decimal:
        return rounding(Decimal(data), 4)


class DatetimeCodec(DefaultCodec):
    """ 日期时间编解码类 """
    type_name = 'datetime'

    def on_load(self, data: Union[str, int, float, datetime]) -> datetime:
        if isinstance(data, str):
            return datetime.strptime(data, datetime_format).astimezone()
        elif isinstance(data, (int, float)):
            return datetime.fromtimestamp(int(data / 1000)).astimezone()
        elif isinstance(data, datetime):
            return data.astimezone()
        raise InvalidValueException(f'{str(data)}无法转换成{self.type_name}类型')

    def on_to_json(self, data: Union[datetime]) -> str:
        if isinstance(data, datetime):
            return data.astimezone().strftime(datetime_format)
        raise InvalidValueException(f'{str(data)}无法转换成{self.type_name}类型')

    def on_to_db(self, data: Union[str, int, float, datetime]) -> datetime:
        return self.on_load(data)


class MongoDB(Database):
    """ MongoDB数据库的实现类 """

    database_type: str = 'mongodb'

    def init(self, model_class, **kwargs):
        for codec_class in (IDCodec, DecimalCodec, DatetimeCodec):
            CodecFactory.set(codec_class, True)
        return self

    def connect(self, uri: str) -> 'Database':
        logging.info(f'连接MongoDB')
        self_codec_options = codec_options.CodecOptions(
            type_registry=codec_options.TypeRegistry([MongoDBDecimalCodec()]),
            tz_aware=True,
        )
        self.client = AsyncIOMotorClient(uri, tz_aware=True)
        self.database = self.client.get_default_database(codec_options=self_codec_options)
        return self

    async def close(self):
        await self.client.close()
        self.client = None
        self.database = None

    def new_query(self, table_name: str) -> Query:
        """ 获取一个新的查询 """
        return MongoDBQuery(table_name)

    def new_operation(self, table_name: str) -> Operation:
        """ 获取一个新的操作 """
        operation = MongoDBOperation(table_name)
        return operation

    async def on_find_one(self, query: Query, **kwargs) -> dict:
        query.filter.update({'$sort': {'_id': 1}, '$skip': 0, '$limit': 1})
        table_name, _, data_pipeline, _, _ = query.get_query()
        collection: AgnosticCollection = self.database[table_name]

        data = await collection.aggregate(data_pipeline, **kwargs).to_list(1)
        return data[0] if (data and isinstance(data, list)) else {}

    async def on_search(self, query: Query, **kwargs) -> Tuple[dict, list]:
        table_name, count_pipeline, data_pipeline, skip, limit = query.get_query()
        collection: AgnosticCollection = self.database[table_name]

        count = await collection.aggregate(count_pipeline, **kwargs).to_list(query.max_length)
        total = count and count[0]['count'] or 0
        data = await collection.aggregate(data_pipeline, **kwargs).to_list(query.max_length)

        return {'skip': skip, 'limit': limit, 'total': total}, data

    async def on_execute(self, operation_list: List[Operation], **kwargs) -> dict:

        bulk_dict = {}
        for op in operation_list:
            if op.operation not in ['insert', 'update', 'delete']:
                continue
            if op.table_name not in bulk_dict:
                bulk_dict[op.table_name] = []
            _table_name, _operation_object = op.get_operation()
            bulk_dict[_table_name].append(_operation_object)

        assert_dict = kwargs.get('assert_dict') or {}
        result_dict = {}
        async with await self.client.start_session() as session:
            session.start_transaction()
            for collection_name, bulk_list in bulk_dict.items():
                collection: AgnosticCollection = self.database[collection_name]
                result = {
                    'insert': 0,
                    'update': 0,
                    'delete': 0,
                }
                for i in range(0, len(bulk_list), 1000):
                    _result = await collection.bulk_write(bulk_list[i:i + 1000], session=session)
                    result['insert'] += _result.bulk_api_result['nInserted']
                    result['update'] += _result.bulk_api_result['nModified']
                    result['delete'] += _result.bulk_api_result['nRemoved']

                result_dict[collection_name] = result

                # 检查断言
                _assert_dict = assert_dict.get(collection_name)
                if _assert_dict and isinstance(_assert_dict, dict):
                    for i in ['insert', 'update', 'delete']:
                        expected_value = _assert_dict.get(i)
                        if expected_value is None:
                            continue
                        if expected_value != result[i]:
                            await session.abort_transaction()
                            message = f'[{collection_name}]表的[{i}]操作预期行数[{expected_value}],实际行数[{result[i]}]'
                            raise Exception(message)

            await session.commit_transaction()

        return result_dict

    async def init_index(self, operation_list: List[Operation]) -> dict:
        """ 初始化索引 """

        bulk_dict = {}
        for op in operation_list:
            if op.operation != 'index':
                continue
            if op.table_name not in bulk_dict:
                bulk_dict[op.table_name] = []
            _table_name, _operation_object = op.get_operation()
            bulk_dict[_table_name].append(_operation_object)

        result_dict = {}
        for collection_name, bulk_list in bulk_dict.items():
            collection: AgnosticCollection = self.database[collection_name]
            result_dict[collection_name] = await collection.create_indexes(bulk_list)

        return result_dict
